import relibmss as ms

# Create a binary decision diagram
bss = ms.BSS()

# Define variables
A = bss.defvar('A')
B = bss.defvar('B')
C = bss.defvar('C')

# Make a tree
top = A & B | C

# Draw the BDD
bdd = bss.getbdd(top)
source = bdd.dot() # source is a string of the dot language
print(source)

# Example: Display the BDD in Jupyter Notebook
from graphviz import Source
Source(source)



import relibmss as ms

# Define gates
def gate1(mss, x, y):
    return mss.switch([
        mss.case(cond=mss.And([x == 0, y == 0]), then=0),
        mss.case(cond=mss.Or([x == 0, y == 0]), then=1),
        mss.case(cond=mss.Or([x == 2, y == 2]), then=3),
        mss.case(cond=None, then=2) # default
    ])

def gate2(mss, x, y):
    return mss.switch([
        mss.case(cond=x == 0, then=0),
        mss.case(cond=None, then=y)
    ])

mss = ms.MSS()

A = mss.defvar('A', 2)
B = mss.defvar('B', 3)
C = mss.defvar('C', 3)

sx = gate1(mss, B, C)
ss = gate2(mss, A, sx)

mdd = mss.minpath(ss)
print(mdd.dot())



import relibmss as ms

# Create a binary system (fault tree)
bss = ms.BSS()

# Define events (This version only supports repeated events)
A = bss.defvar('A')
B = bss.defvar('B')
C = bss.defvar('C')

# Make a tree
top = A & B | C # & is AND gate, | is OR gate

# Set probabilities
prob = {
    'A': 0.1,
    'B': 0.2,
    'C': 0.3
}

# Calculate the probability
print(bss.prob(top, prob))

# Set the interval of the probability
prob = {
    'A': (0.1, 0.2),
    'B': (0.2, 0.3),
    'C': (0.3, 0.4)
}

# Calculate the probability
print(bss.prob_interval(top, prob))



import relibmss as ms

# Define gates
def gate1(mss, x, y):
    return mss.switch([
        mss.case(cond=mss.And([x == 0, y == 0]), then=0),
        mss.case(cond=mss.Or([x == 0, y == 0]), then=1),
        mss.case(cond=mss.Or([x == 2, y == 2]), then=3),
        mss.case(cond=None, then=2) # default
    ])

def gate2(mss, x, y):
    return mss.switch([
        mss.case(cond=x == 0, then=0),
        mss.case(cond=None, then=y)
    ])

mss = ms.MSS() # Context for the multi-state system

# Define variables

A = mss.defvar('A', 2) # 2 states
B = mss.defvar('B', 3) # 3 states
C = mss.defvar('C', 3) # 3 states

# Define a multi-state system
sx = gate1(mss, B, C)
ss = gate2(mss, A, sx)

# Define probabilities
prob = {
    'A': [0.1, 0.9],
    'B': [0.2, 0.3, 0.5],
    'C': [0.3, 0.4, 0.3]
}

# Calculate the probability
print(mss.prob(ss, prob, [0,1,2]))



import relibmss as ms

# Create a binary system (fault tree)
bss = ms.BSS()

# Define events (This version only supports repeated events)
A = bss.defvar('A')
B = bss.defvar('B')
C = bss.defvar('C')

# Make a tree
top = A & B | C # & is AND gate, | is OR gate

# Set probabilities
prob = {
    'A': 0.1,
    'B': 0.2,
    'C': 0.3
}

# Calculate the probability
print(bss.prob(top, prob))

# top = 1-(1-pa*pb)*(1-pc) = pa*pb+pc-pa*pb*pc
# top / pa = pb - pb*pc = 0.2 - 0.2*0.3 = 0.14
# top / pb = pa - pa*pc = 0.1 - 0.1*0.3 = 0.07
# top / pc = 1 - pa*pb = 1 - 0.1*0.2 = 0.98

print(bss.bmeas(top, prob))

# Set the interval of the probability
prob = {
    'A': (0.1, 0.2),
    'B': (0.2, 0.3),
    'C': (0.3, 0.4)
}

# Calculate the probability
print(bss.prob_interval(top, prob))

print(bss.bmeas_interval(top, prob))

###

prob = {
    'A': 0.5,
    'B': 0.5,
    'C': 0.5
}

# structure importance measure
print(bss.bmeas(top, prob))



import relibmss as ms

# Create a binary system
bss = ms.BSS()

# Define events (This version only supports repeated events)
A = bss.defvar('A')
B = bss.defvar('B')
C = bss.defvar('C')

# Make a system
top = bss.kofn(2, [A, B, C]) # k-of-n gate

# Convert the ZDD representation to a list of sets
path = bss.getbdd(top).bdd_extract([True])
print('All paths which is to be one')
for x in path:
    print(x)

# Obtain the minimal path vectors
s = bss.minpath(top) # s is a set of minimal path vectors (ZDD representation)

# Convert the ZDD representation to a list of sets
min_path = s.zdd_extract([True])
print('The number of minimal path vectors:', len(min_path))
for x in min_path:
    print(x)



# 必要なライブラリのインポート
import numpy as np
import relibmss as ms
from functools import reduce

# def for a gate with switch-case structure
def max_gate(mss, x, y):
  return mss.switch([
        mss.case(cond= x < y, then=y),
        mss.case(cond=None, then=x) # default
    ])

def min_gate(mss, x, y):
  return mss.switch([
        mss.case(cond= x < y, then=x),
        mss.case(cond=None, then=y) # default
    ])

def var2(mss, x):
  return mss.switch([
        mss.case(cond= x == 0, then=0),
        mss.case(cond=None, then=x-1) # default
    ])

def maxs(mss, x):
  return reduce(lambda a, b: max_gate(mss, a, b), x, mss.const(0))

def mins(mss, x):
  return reduce(lambda a, b: min_gate(mss, a, b), x, mss.const(2))

# 基地局を表すクラスの定義

class Base:
  def __init__(self, name, p, r1, r2):
    # name: ラベル，p(x, y); 座標，r1: 範囲, r2 (r2 > r1)
    self.name = name
    self.p = p
    self.r1 = r1
    self.r2 = r2

  def __repr__(self) -> str:
    return f"Base({self.name}, {self.p}, {self.r1}, {self.r2})"

  def check_dist1(self, p):
    # 与えられた座標p(x,y)が範囲内かどうかを判定する
    # pはnumpy array
    return np.linalg.norm(self.p - p) <= self.r1

  def check_dist2(self, p):
    # 与えられた座標p(x,y)が範囲内かどうかを判定する
    # pはnumpy array
    return self.r1 < np.linalg.norm(self.p - p) <= self.r2

def get_bases(base_list, p):
  # 与えられた座標p(x,y)をカバーする基地局のリストを返す
  bases1 = [b for b in base_list if b.check_dist1(p)]
  bases2 = [b for b in base_list if b.check_dist2(p)]
  return (bases1, bases2)

def get_bases2(base_list, ps):
  # 与えられた座標集合ps={p1, p2, ...}が作るエリアをカバーする基地局のリストを返す
  #    全部の点をカバーしていればエリアをカバーすることになる
  bases1 = [b for b in base_list if all([b.check_dist1(p) for p in ps])]
  bases2 = [b for b in base_list if all([b.check_dist1(p) or b.check_dist2(p) for p in ps])]
  bases2 = [b for b in bases2 if b not in bases1]
  return (bases1, bases2)

np.random.seed(12247) # 乱数シードの設定
r1 = 0.5 # カバーする半径（この場合は全部の基地が同じ半径）
r2 = 1.0 # カバーする半径（この場合は全部の基地が同じ半径）
n = 10 # 基地数
base_list = [Base('base'+str(i), np.random.rand(2), r1, r2) for i in range(n)] # ランダムに[0,1] x [0,1]エリアに基地をn個配置

mss = ms.MSS()
vars = {b.name: mss.defvar(b.name, 3) for b in base_list} # 基地に対応する変数の作成
prob = {b.name: [0.01, 0.3, 0.69] for b in base_list} # 各基地が故障しない確率

orders = {x[1][0]:x[0] for x in enumerate(sorted([(b.name, np.linalg.norm(b.p)) for b in base_list], key=lambda x: x[1]))}
print(orders)
mss.set_varorder(orders)

gn = 5 # グリッド数 gn x gn 個の点を均等に配置
grid_x = np.linspace(0, 1, gn)
grid_y = np.linspace(0, 1, gn)
# 点座標毎 [[0,0], [0,0.1], ...] のようなデータ
grid = [np.array([x, y]) for x in grid_x for y in grid_y]
# エリア毎の座標集合のデータ [[[0,0], [0,0.1], [0.1, 0]], [[0,0.1], [0.1,0], [0.1,0.1]], ...]
grid2 = [np.array([[grid_x[xi], grid_y[yi]], [grid_x[xi], grid_y[yi-1]], [grid_x[xi-1], grid_y[yi]], [grid_x[xi-1], grid_y[yi-1]]]) for xi in range(1, gn) for yi in range(1, gn)]

def make_tree(mss, g, base_list):
  bases1 = [vars[b.name] for b in get_bases(base_list, g)[0]]
  bases2 = [vars[b.name] for b in get_bases(base_list, g)[1]]
  return max_gate(mss, maxs(mss, bases1), var2(mss, maxs(mss, bases2)))

xs1 = [make_tree(mss, g, base_list) for g in grid]
sy1 = mins(mss, xs1)

def make_tree2(mss, gs, base_list):
  bases1 = [vars[b.name] for b in get_bases2(base_list, gs)[0]]
  bases2 = [vars[b.name] for b in get_bases2(base_list, gs)[1]]
  return max_gate(mss, maxs(mss, bases1), var2(mss, maxs(mss, bases2)))

xs2 = [make_tree2(mss, gs, base_list) for gs in grid2]
sy2 = mins(mss, xs2)

print(xs1[0].to_rpn())

print(sy1.to_rpn())

mdd = mss.getmdd(sy1)

print(mdd.dot())



import relibmss as ms

def _to_rpn(expr):
    stack = [expr]
    rpn = []
    cache = set([])
    while len(stack) > 0:
        node = stack.pop()
        idnum = str(id(node))
        if isinstance(node, tuple) and node[0] == "save" and len(stack) > 0:
            idnum = node[1]
            rpn.append('save({})'.format(idnum))
            cache.add(idnum)
        elif isinstance(node, tuple) and node[0] == "save":
            pass
        elif idnum in cache:
            rpn.append('load({})'.format(idnum))
        elif isinstance(node.value, tuple):
            stack.append(("save", idnum))
            for i in range(len(node.value) - 1, -1, -1):
                stack.append(node.value[i])
        else:
            rpn.append(str(node.value))
    return ' '.join(rpn)

class _Expression:
    def __init__(self, value):
        self.value = value

    def __and__(self, other):
        if not isinstance(other, _Expression):
            other = _Expression(other)
        return _Expression((self, other, _Expression('&')))
    
    def __or__(self, other):
        if not isinstance(other, _Expression):
            other = _Expression(other)
        return _Expression((self, other, _Expression('|')))

    def __str__(self):
        if isinstance(self.value, tuple):
            return _to_rpn(self)
        return str(self.value)

    def _to_rpn(self):
        if isinstance(self.value, tuple):
            return _to_rpn(self)
        return str(self.value)

class Context:
    def __init__(self):
        self.vars = set([])
        self.bdd = ms.BDD()

    def defvar(self, name):
        self.vars.add(name)
        return _Expression(name)
    
    def get_varorder(self):
        return self.bdd.get_varorder()

    def set_varorder(self, x: list):
        for varname in x:
            self.bdd.defvar(varname)

    def __str__(self):
        return str(self.vars)
    
    def getbdd(self, arg: _Expression):
        if not isinstance(arg, _Expression):
            arg = _Expression(arg)
        rpn = arg._to_rpn()
        return self.bdd.rpn(rpn)
    
    def const(self, value):
        return _Expression(value)

    def And(self, args: list):
        assert len(args) > 0
        if not isinstance(args[0], _Expression):
            args[0] = _Expression(args[0])
        if len(args) == 1:
            return args[0]
        x = args[0]
        for y in args[1:]:
            if not isinstance(y, _Expression):
                y = _Expression(y)
            x = _Expression((x, y, _Expression('&')))
        return x

    def Or(self, args: list):
        assert len(args) > 0
        if not isinstance(args[0], _Expression):
            args[0] = _Expression(args[0])
        if len(args) == 1:
            return args[0]
        x = args[0]
        for y in args[1:]:
            if not isinstance(y, _Expression):
                y = _Expression(y)
            x = _Expression((x, y, _Expression('|')))
        return x

    def Not(self, arg: _Expression):
        if not isinstance(arg, _Expression):
            arg = _Expression(arg)
        return _Expression((arg, _Expression('!')))

    def ifelse(self, condition: _Expression, then_expr: _Expression, else_expr: _Expression):
        if not isinstance(condition, _Expression):
            condition = _Expression(condition)
        if not isinstance(then_expr, _Expression):
            then_expr = _Expression(then_expr)
        if not isinstance(else_expr, _Expression):
            else_expr = _Expression(else_expr)
        return _Expression((condition, then_expr, else_expr, _Expression('?')))

    def kofn(self, k: int, args: list):
        assert k <= len(args)
        if k == 1:
            return self.Or(args)
        elif k == len(args):
            return self.And(args)
        else:
            return self.ifelse(args[0], self.kofn(k-1, args[1:]), self.kofn(k, args[1:]))

    def prob(self, arg: _Expression, values: dict, ss = [True]):
        top = self.getbdd(arg)
        return top.prob(values, ss)
    
    def bmeas(self, arg: _Expression, values: dict, ss = [True]):
        top = self.getbdd(arg)
        return top.bmeas(values, ss)

    def prob_interval(self, arg: _Expression, values: dict, ss = [True]):
        values = {k: ms.Interval(v[0], v[1]) for k, v in values.items()}
        top = self.getbdd(arg)
        return top.prob_interval(values, ss)

    def bmeas_interval(self, arg: _Expression, values: dict, ss = [True]):
        values = {k: ms.Interval(v[0], v[1]) for k, v in values.items()}
        top = self.getbdd(arg)
        return top.bmeas_interval(values, ss)

    def minpath(self, arg: _Expression):
        top = self.getbdd(arg)
        return top.minpath()
    




import relibmss as ms

def test_mss():
    ctx = ms.MSS()
    x = ctx.defvar("x", 2)
    y = ctx.defvar("y", 3)
    z = ctx.defvar("z", 3)
    v = x * y + z
    v = ctx.And([x >= 1, y <= 1, z == 0])
    print(v)

    tree = ctx.mdd.rpn("x 1 >= y 1 <= &&", ctx.vars)
    tree2 = ctx.getmdd(v)

    print(tree.dot())
    print(tree2.dot())

def test_mss3():
    mss = ms.MSS()
    x = mss.defvar("x", 3)
    y = mss.defvar("y", 3)
    z = mss.defvar("z", 3)
    s1 = mss.ifelse(x + y == z, 100, 200)
    print(s1)
    tree = mss.getmdd(s1)
    print(tree.dot())

def test_mdd():
    mdd = ms.MDD()
    x = mdd.defvar("x", 3)
    y = mdd.defvar("y", 3)
    z = mdd.defvar("z", 3)
    # tree = mdd.rpn("x y + z == 100 200 ?", {})
    # print(tree.dot())

def test_bdd1():
    bdd = ms.BDD()
    x = bdd.defvar("x")
    y = bdd.defvar("y")
    z = bdd.defvar("z")
    v = x & y | z
    print(v)
    print(v.dot())

def test_mdd2():
    mdd = ms.MDD()
    x = mdd.defvar("x", 3)
    y = mdd.defvar("y", 3)
    z = mdd.defvar("z", 3)
    v = x + y == z + 1
    print(v.dot())
    v2 = mdd.ifelse(x + y == z, 100, 200)
    print(v2.dot())
    v3 = mdd.ifelse(mdd.And([x + y == z, x == z]), 100, 200)
    print(v3.dot())

def test_mdd3():
    mdd = ms.MDD()
    x = mdd.defvar("x", 3)
    y = mdd.defvar("y", 3)
    z = mdd.defvar("z", 3)
    v = mdd.ifelse(x + y == z, 100, 200)
    print(v.dot())

def test_mdd4():
    mdd = ms.MDD()
    x = mdd.defvar("x", 3)
    y = mdd.defvar("y", 3)
    z = mdd.defvar("z", 3)
    v = mdd.ifelse(mdd.And([x + y == z, x == z]), 100, 200)
    print(v.dot())

def test_mdd5():
    mdd = ms.MDD()
    x = mdd.defvar("x", 3)
    y = mdd.defvar("y", 3)
    z = mdd.defvar("z", 3)
    v = mdd.ifelse(mdd.Or([x + y == z, x == z]), 100, 200)
    print(v.dot())

def test_mdd5():
    mdd = ms.MDD()
    x = mdd.defvar("x", 10)
    y = mdd.defvar("y", 3)
    z = mdd.defvar("z", 3)
    v = mdd.ifelse(mdd.Not(mdd.Or([x + y == z, x == z])), 100, 200)
    print(v.dot())

def test_mss6():
    ctx = ms.MSS()
    A = ctx.defvar("A", 2)
    B = ctx.defvar("B", 3)
    C = ctx.defvar("C", 3)

    def gate1(x, y):
        return ctx.ifelse(
            ctx.And([x == 0, y == 0]), 0,
            ctx.ifelse(
                ctx.Or([x == 0, y == 0]), 1,
                    ctx.ifelse(
                        ctx.Or([x == 2, y == 2]), 3,
                    2
                )
            )
        )
    
    def gate2(x, y):
        return ctx.ifelse(x == 0, 0, y)

    ctx.mdd.defvar("C", 3)
    ctx.mdd.defvar("B", 3)
    ctx.mdd.defvar("A", 2)

    sx = gate1(B, C)
    ss = gate2(A, sx)

    print(ss)
    mdd = ctx.getmdd(ss)
    print(mdd.dot())

def test_mss7():
    ctx = ms.MSS()
    A = ctx.defvar("A", 2)
    B = ctx.defvar("B", 3)
    C = ctx.defvar("C", 3)

    x = ctx.switch([
        ctx.case(cond=ctx.And([A == 0, B == 0]), then=0),
        ctx.case(cond=ctx.Or([A == 0, B == 0]), then=1),
        ctx.case(cond=None, then=2)
    ])

    print(x)

def test_mss8():
    ctx = ms.MSS()
    A = ctx.defvar("A", 2)
    B = ctx.defvar("B", 3)
    C = ctx.defvar("C", 3)
    ctx.set_varorder({"A": 2, "B": 1, "C": 0}) # this should be done before making MDD

    def gate1(x, y):
        return ctx.switch([
            ctx.case(cond=ctx.And([x == 0, y == 0]), then=0),
            ctx.case(cond=ctx.Or([x == 0, y == 0]), then=1),
            ctx.case(cond=ctx.Or([x == 2, y == 2]), then=3),
            ctx.case(cond=None, then=2)
        ])

    def gate2(x, y):
        return ctx.switch([
            ctx.case(cond=x == 0, then=0),
            ctx.case(cond=None, then=y)
        ])    

    sx = gate1(B, C)
    ss = gate2(A, sx)

    print(ss)
    mdd = ctx.getmdd(ss) # this is the time when MDD is created
    print(mdd.dot())

def test_mss9():
    ctx = ms.MSS()
    A = ctx.defvar("A", 2)
    B = ctx.defvar("B", 3)
    C = ctx.defvar("C", 3)
    ctx.set_varorder({"A": 2, "B": 1, "C": 0}) # this should be done before making MDD

    def gate1(x, y):
        return ctx.switch([
            ctx.case(cond=ctx.And([x == 0, y == 0]), then=0),
            ctx.case(cond=ctx.Or([x == 0, y == 0]), then=1),
            ctx.case(cond=ctx.Or([x == 2, y == 2]), then=3),
            ctx.case(cond=None, then=2)
        ])

    def gate2(x, y):
        return ctx.switch([
            ctx.case(cond=x == 0, then=0),
            ctx.case(cond=None, then=y)
        ])    

    sx = gate1(B, C)
    ss = gate2(A, sx)

    prob = {
        "A": [0.9, 0.1],
        "B": [0.5, 0.2, 0.3],
        "C": [0.2, 0.3, 0.5]
    }

    mdd = ctx.getmdd(ss) # this is the time when MDD is created
    print(ctx.prob(ss, prob))

def test_mss10():
    ctx = ms.MSS()
    A = ctx.defvar("A", 2)
    B = ctx.defvar("B", 3)
    C = ctx.defvar("C", 3)
    ctx.set_varorder({"A": 2, "B": 1, "C": 0}) # this should be done before making MDD

    def gate1(x, y):
        return ctx.switch([
            ctx.case(cond=ctx.And([x == 0, y == 0]), then=0),
            ctx.case(cond=ctx.Or([x == 0, y == 0]), then=1),
            ctx.case(cond=ctx.Or([x == 2, y == 2]), then=3),
            ctx.case(cond=None, then=2)
        ])

    def gate2(x, y):
        return ctx.switch([
            ctx.case(cond=x == 0, then=0),
            ctx.case(cond=None, then=y)
        ])    

    sx = gate1(B, C)
    ss = gate2(A, sx)

    prob = {
        "A": [(0.9, 0.91), (0.09, 0.1)],
        "B": [(0.5, 0.51), (0.2, 0.21), (0.28, 0.3)],
        "C": [(0.2, 0.21), (0.3, 0.31), (0.5, 0.51)]
    }

    mdd = ctx.getmdd(ss) # this is the time when MDD is created
    print(ctx.prob_interval(ss, prob))

def test_mss11():
    ctx = ms.MSS()
    A = ctx.defvar("A", 2)
    B = ctx.defvar("B", 3)
    C = ctx.defvar("C", 3)
    ctx.set_varorder({"A": 2, "B": 1, "C": 0}) # this should be done before making MDD

    def gate1(x, y):
        return ctx.switch([
            ctx.case(cond=ctx.And([x == 0, y == 0]), then=0),
            ctx.case(cond=ctx.Or([x == 0, y == 0]), then=1),
            ctx.case(cond=ctx.Or([x == 2, y == 2]), then=3),
            ctx.case(cond=None, then=2)
        ])

    def gate2(x, y):
        return ctx.switch([
            ctx.case(cond=x == 0, then=0),
            ctx.case(cond=None, then=y)
        ])    

    sx = gate1(B, C)
    ss = gate2(A, sx)

    mdd = ctx.getmdd(ss) # this is the time when MDD is created
    v = mdd.mpvs()
    print(v.dot())

def test_mss12():
    ctx = ms.MSS()
    A = ctx.defvar("A", 2)
    B = ctx.defvar("B", 3)
    C = ctx.defvar("C", 3)

    def gate1(x, y):
        return ctx.switch([
            ctx.case(cond=ctx.And([x == 0, y == 0]), then=0),
            ctx.case(cond=ctx.Or([x == 0, y == 0]), then=1),
            ctx.case(cond=ctx.Or([x == 2, y == 2]), then=3),
            ctx.case(cond=None, then=2)
        ])

    def gate2(x, y):
        return ctx.switch([
            ctx.case(cond=x == 0, then=0),
            ctx.case(cond=None, then=y)
        ])    

    sx = gate1(B, C)
    ss = gate2(A, sx)

    mdd = ctx.getmdd(ss) # this is the time when MDD is created
    v = mdd.mpvs()
    print(v.dot())

def test_mss13():
    def gate1(ctx, x, y):
        return ctx.switch([
            ctx.case(cond=ctx.And([x == 0, y == 0]), then=0),
            ctx.case(cond=ctx.Or([x == 0, y == 0]), then=1),
            ctx.case(cond=ctx.Or([x == 2, y == 2]), then=3),
            ctx.case(cond=None, then=2)
        ])

    # def gate2(ctx, x, y):
    #     return ctx.switch([
    #         ctx.case(cond=x == 0, then=0),
    #         ctx.case(cond=None, then=y)
    #     ])

    ctx = ms.MSS()
    x = ctx.defvar("x", 3)
    y = ctx.defvar("y", 3)
    xdash = ctx.defvar("xdash", 3)
    ydash = ctx.defvar("ydash", 3)

    # gate1 is the increasing function or not?
    ss = ctx.ifelse(ctx.And([x <= xdash, y <= ydash]), gate1(ctx, x, y) <= gate1(ctx, xdash, ydash), True)
    print(ss)

    v = ctx.getmdd(ss)
    print(v.dot())

def test_mss13():
    ctx = ms.MSS()
    A = ctx.defvar("A", 2)
    B = ctx.defvar("B", 3)
    C = ctx.defvar("C", 3)

    def gate1(x, y):
        return ctx.switch([
            ctx.case(cond=ctx.And([x == 0, y == 0]), then=0),
            ctx.case(cond=ctx.Or([x == 0, y == 0]), then=1),
            ctx.case(cond=ctx.Or([x == 2, y == 2]), then=3),
            ctx.case(cond=None, then=2)
        ])

    def gate2(x, y):
        return ctx.switch([
            ctx.case(cond=x == 0, then=0),
            ctx.case(cond=None, then=y)
        ])    

    sx = gate1(B, C)
    ss = gate2(A, sx)

    mdd = ctx.getmdd(ss) # this is the time when MDD is created
    v = mdd.mpvs()
    print(v.dot())

def test_mss14():
    ctx = ms.MSS()
    A = ctx.defvar("A", 2)
    B = ctx.defvar("B", 3)
    C = ctx.defvar("C", 3)

    def gate1(x, y):
        return ctx.switch([
            ctx.case(cond=ctx.And([x == 0, y == 0]), then= x == y),
            ctx.case(cond=ctx.Or([x == 0, y == 0]), then=x != y),
            ctx.case(cond=ctx.Or([x == 2, y == 2]), then=x <= y),
            ctx.case(cond=None, then=False)
        ])

    sx = gate1(B, C)

    mdd = ctx.getmdd(sx)
    v = mdd.mpvs()
    print(v.dot())

    prob = {
        "B": [0.5, 0.5, 0.0],
        "C": [0.5, 0.2, 0.3]
    }

    print(ctx.prob(sx, prob))



# Import classes and rename them for clarity
from .relibmss import PyBddMgr as BDD
from .relibmss import PyMddMgr as MDD
from .relibmss import PyBddNode as BddNode
from .relibmss import PyMddNode as MddNode
from .relibmss import Interval
from .mss import Context as MSS
from .bss import Context as BSS

# Define what should be exposed when `from relibmss import *` is used
__all__ = ["BddNode", "BDD", "MddNode", "MDD", "MSS", "BSS", "Interval"]



import relibmss as ms

# def _to_rpn(expr):
#     stack = [expr]
#     rpn = []
#     while len(stack) > 0:
#         node = stack.pop()
#         if isinstance(node.value, tuple):
#             for i in range(len(node.value) - 1, -1, -1):
#                 stack.append(node.value[i])
#         else:
#             rpn.append(str(node.value))
#     return ' '.join(rpn)

def _to_rpn(expr):
    stack = [expr]
    rpn = []
    cache = set([])
    while len(stack) > 0:
        node = stack.pop()
        idnum = str(id(node))
        if isinstance(node, tuple) and node[0] == "save" and len(stack) > 0:
            idnum = node[1]
            rpn.append('save({})'.format(idnum))
            cache.add(idnum)
        elif isinstance(node, tuple) and node[0] == "save":
            pass
        elif idnum in cache:
            rpn.append('load({})'.format(idnum))
        elif isinstance(node.value, tuple):
            stack.append(("save", idnum))
            for i in range(len(node.value) - 1, -1, -1):
                stack.append(node.value[i])
        else:
            rpn.append(str(node.value))
    return ' '.join(rpn)

class _Expression:
    def __init__(self, value):
        self.value = value

    def __add__(self, other):
        if not isinstance(other, _Expression):
            other = _Expression(other)
        return _Expression((self, other, _Expression('+')))
    
    def __sub__(self, other):
        if not isinstance(other, _Expression):
            other = _Expression(other)
        return _Expression((self, other, _Expression('-')))
    
    def __mul__(self, other):
        if not isinstance(other, _Expression):
            other = _Expression(other)
        return _Expression((self, other, _Expression('*')))
    
    def __truediv__(self, other):
        if not isinstance(other, _Expression):
            other = _Expression(other)
        return _Expression((self, other, _Expression('/')))
    
    def __eq__(self, other):
        if not isinstance(other, _Expression):
            other = _Expression(other)
        return _Expression((self, other, _Expression('==')))
    
    def __ne__(self, other):
        if not isinstance(other, _Expression):
            other = _Expression(other)
        return _Expression((self, other, _Expression('!=')))
    
    def __lt__(self, other):
        if not isinstance(other, _Expression):
            other = _Expression(other)
        return _Expression((self, other, _Expression('<')))
    
    def __le__(self, other):
        if not isinstance(other, _Expression):
            other = _Expression(other)
        return _Expression((self, other, _Expression('<=')))
    
    def __gt__(self, other):
        if not isinstance(other, _Expression):
            other = _Expression(other)
        return _Expression((self, other, _Expression('>')))
    
    def __ge__(self, other):
        if not isinstance(other, _Expression):
            other = _Expression(other)
        return _Expression((self, other, _Expression('>=')))
    
    def __str__(self):
        if isinstance(self.value, tuple):
            return _to_rpn(self)
        return str(self.value)

    def to_rpn(self):
        if isinstance(self.value, tuple):
            return _to_rpn(self)
        return str(self.value)

class _Case:
    def __init__(self, cond, then):
        self.cond = cond
        self.then = then

class Context:
    def __init__(self):
        self.vars = {}
        self.mdd = ms.MDD()

    def defvar(self, name, domain):
        self.vars[name] = domain
        return _Expression(name)
    
    def set_varorder(self, x: dict):
        for varname in sorted(x, key=x.get):
            self.mdd.defvar(varname, self.vars[varname])
    
    def __str__(self):
        return str(self.vars)
    
    def getmdd(self, arg: _Expression):
        if not isinstance(arg, _Expression):
            arg = _Expression(arg)
        rpn = arg.to_rpn()
        return self.mdd.rpn(rpn, self.vars)
    
    def const(self, value):
        return _Expression(value)

    def And(self, args: list):
        assert len(args) > 0
        if not isinstance(args[0], _Expression):
            args[0] = _Expression(args[0])
        if len(args) == 1:
            return args[0]
        x = args[0]
        for y in args[1:]:
            if not isinstance(y, _Expression):
                y = _Expression(y)
            x = _Expression((x, y, _Expression('&&')))
        return x

    def Or(self, args: list):
        assert len(args) > 0
        if not isinstance(args[0], _Expression):
            args[0] = _Expression(args[0])
        if len(args) == 1:
            return args[0]
        x = args[0]
        for y in args[1:]:
            if not isinstance(y, _Expression):
                y = _Expression(y)
            x = _Expression((x, y, _Expression('||')))
        return x

    def Not(self, arg: _Expression):
        if not isinstance(arg, _Expression):
            arg = _Expression(arg)
        return _Expression((arg, _Expression('!')))

    def ifelse(self, condition: _Expression, then_expr: _Expression, else_expr: _Expression):
        if not isinstance(condition, _Expression):
            condition = _Expression(condition)
        if not isinstance(then_expr, _Expression):
            then_expr = _Expression(then_expr)
        if not isinstance(else_expr, _Expression):
            else_expr = _Expression(else_expr)
        return _Expression((condition, then_expr, else_expr, _Expression('?')))
    
    def case(self, then, cond = None):
        if not isinstance(cond, _Expression):
            cond = _Expression(cond)
        if not isinstance(then, _Expression):
            then = _Expression(then)
        return _Case(cond=cond, then=then)
    
    def switch(self, conds: list):
        assert len(conds) >= 2
        if len(conds) == 2:
            assert isinstance(conds[0], _Case) and isinstance(conds[1], _Case)
            return self.ifelse(conds[0].cond, conds[0].then, conds[1].then)
        else:
            x = conds[0]
            if not isinstance(x, _Case):
                raise ValueError("The element must be a Case object")
            return self.ifelse(x.cond, x.then, self.switch(conds[1:]))

    def prob(self, arg: _Expression, values: dict, sv: list):
        top = self.getmdd(arg)
        return top.prob(values, sv)
        
    def prob_interval(self, arg: _Expression, values: dict, sv: list):
        values = {k: [ms.Interval(u[0], u[1]) for u in v] for k, v in values.items()}
        top = self.getmdd(arg)
        return top.prob_interval(values, sv)
    
    def minpath(self, arg: _Expression):
        top = self.getmdd(arg)
        return top.minpath()






import relibmss as ms
import numpy as np

class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y
    
    def dist(self, p):
        return np.sqrt((self.x - p.x)**2 + (self.y - p.y)**2)

np.random.seed(1234)

bss = ms.BSS()
m = 50
n = 100
ps = {'p'+str(i): Point(np.random.rand(), np.random.rand()) for i in range(m)}
vars = {'p'+str(i): bss.defvar('p'+str(i)) for i in range(m)}
# sort ps based on the distance from (0,0)
sortedps = [k for (k,v) in sorted(ps.items(), key=lambda x: x[1].dist(Point(0,0)))]
print(sortedps)
bss.set_varorder(sortedps)

grid = {'grid_{}_{}'.format(i,j): Point(x,y) for (i,x) in enumerate(np.linspace(0,1,n)) for (j,y) in enumerate(np.linspace(0,1,n))}

r = 0.5
result = []
for (k,v) in grid.items():
    nm = [pn for (pn, pv) in ps.items() if pv.dist(v) <= r]
    tmp = bss.const(False)
    for x in nm:
        tmp = vars[x] | tmp
    result.append(tmp)

expr = bss.const(True)
for x in result:
    expr = x & expr

# print(str(expr))

bdd = bss.getbdd(expr)
print(bdd.size())

paths = bdd.minpath()
print(paths.zdd_count([True]))

prob = {'p'+str(i): 0.9 for i in range(m)}

print(bdd.prob(prob, [True]))



import relibmss as ms

def test_ft1():
    ctx = ms.BSS()
    x = ctx.defvar("x")
    y = ctx.defvar("y")
    z = ctx.defvar("z")
    v = x & y | z
    u = ctx.kofn(2, [x, y, z])
    print(u)
    print(ctx.getbdd(u).dot())

def test_bdd1():
    bdd = ms.BDD()
    x = bdd.defvar("x")
    y = bdd.defvar("y")
    z = bdd.defvar("z")
    v = x & y | z
    print(v)
    print(v.dot())

def test_ft3():
    ctx = ms.BSS()
    x = ctx.defvar("x")
    y = ctx.defvar("y")
    z = ctx.defvar("z")
    u = ctx.kofn(2, [x, y, z])
    print(u)
    print(ctx.getbdd(u).dot())
    print("prob:", ctx.prob(u, {"x": 0.3, "y": 0.2, "z": 0.1}))
    m = ctx.mpvs(u)
    print('mcs: ', m.extract())

def test_interval4():
    x = ms.Interval(0, 1)
    print(x)

def test_interval5():
    ctx = ms.BSS()
    x = ctx.defvar("x")
    y = ctx.defvar("y")
    z = ctx.defvar("z")
    u = ctx.kofn(2, [x, y, z])
    print(u)
    print(ctx.getbdd(u).dot())
    problist = {
        "x": (1.0e-3, 1.0e-2),
        "y": (1.0e-4, 1.0e-3),
        "z": (1.0e-3, 1.0e-2)
    }
    print("prob:", ctx.prob_interval(u, problist))

